
from .datastructures import *
from .utilities import *
from . import sfunc

__author__ = "Shervin Azadi, and Pirouz Nourian"
__copyright__ = "???"
__credits__ = ["Shervin Azadi", "Pirouz Nourian"]
__license__ = "???"
__version__ = "0.0.2"
__maintainer__ = "Shervin Azadi"
__email__ = "shervinazadi93@gmail.com"
__status__ = "Dev"




"""
Main Data Structure of the topoGenesis
"""

import numpy as np
import pyvista as pv
import itertools
import concurrent.futures
import warnings
import os

__author__ = "Shervin Azadi, and Pirouz Nourian"
__copyright__ = "???"
__credits__ = ["Shervin Azadi", "Pirouz Nourian"]
__license__ = "???"
__version__ = "0.0.2"
__maintainer__ = "Shervin Azadi"
__email__ = "shervinazadi93@gmail.com"
__status__ = "Dev"

file_directory = os.path.dirname(os.path.abspath(__file__))


class lattice(np.ndarray):

    def __new__(subtype, bounds, unit=1, dtype=float, buffer=None, offset=0,
                strides=None, order=None, default_value=None):

        # extracting min and max from bound and discrtizing it
        bounds = np.array(bounds)
        minbound = np.rint(bounds[0] / unit).astype(int)
        maxbound = np.rint(bounds[1] / unit).astype(int)
        bounds = np.array([minbound, maxbound])*unit

        # unit np array
        unit = np.array(unit)

        # raise value error if the size of unit is neighter 1 nor the length of the minimum
        if unit.size != 1 and unit.size != minbound.size:
            raise ValueError(
                'the length of unit array needs to be either 1 or equal to the min/max arrays')

        # calculating shape based on bounds and unit
        shape = 1 + maxbound - minbound

        # set default value
        if default_value != None:
            buffer = np.tile(
                default_value, shape)
            #obj = obj * 0 + default_value

        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to lattice.__array_finalize__
        obj = super(lattice, subtype).__new__(subtype, shape, dtype,
                                              buffer, offset, strides,
                                              order)

        # set the  'bounds' attribute
        obj.bounds = bounds
        # set the attribute 'unit' to itself if it has the same size as the minimum,
        # if the size is 1, tile it with the size of minimum vector
        obj.unit = unit if unit.size == minbound.size else np.tile(
            unit, minbound.size)

        # init an empty connectivity
        obj.connectivity = None
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(lattice, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. lattice():
        #    obj is None
        #    (we're in the middle of the lattice.__new__
        #    constructor, and self.bounds will be set when we return to
        #    lattice.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(lattice):
        #    obj is arr
        #    (type(obj) can be lattice)
        # From new-from-template - e.g lattice[:3]
        #    type(obj) is lattice
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'bounds', because this
        # method sees all creation of default objects - with the
        # lattice.__new__ constructor, but also with
        # arr.view(lattice).
        self.bounds = getattr(obj, 'bounds', None)
        self.bounds = getattr(obj, 'bounds', None)
        self.unit = getattr(obj, 'unit', None)
        self.connectivity = getattr(obj, 'connectivity', None)
        # We do not need to return anything

    @property
    def minbound(self):
        return self.bounds[0]

    @property
    def maxbound(self):
        return self.bounds[1]

    @property
    def centroids(self):
        # extract the indices of the True values # with sparse matrix we don't need to search
        point_array = np.argwhere(self == True)
        # convert to float
        point_array = point_array.astype(float)
        # multiply by unit
        point_array *= self.unit
        # move to minimum
        point_array += self.minbound
        # return as a point cloud
        return cloud(point_array, dtype=float)

    def fast_vis(self, plot, show_outline=True, show_centroids=True):

        # Set the grid dimensions: shape + 1 because we want to inject our values on the CELL data
        grid = pv.UniformGrid()
        grid.dimensions = np.array(self.shape) + 1
        # The bottom left corner of the data set
        grid.origin = self.minbound - self.unit * 0.5
        grid.spacing = self.unit  # These are the cell sizes along each axis
        # Add the data values to the cell data
        grid.cell_arrays["values"] = self.flatten(
            order="F").astype(float)  # Flatten the array!
        # filtering the voxels
        threshed = grid.threshold([0.9, 1.1])

        # adding the voxels: light red
        plot.add_mesh(threshed, show_edges=True, color="#ff8fa3",
                      opacity=0.3, label="Cells")

        if show_outline:
            # adding the boundingbox wireframe
            plot.add_mesh(grid.outline(), color="grey", label="Domain")

        if show_centroids:
            # adding the voxel centeroids: red
            plot.add_mesh(pv.PolyData(self.centroids), color='#ff244c', point_size=5,
                          render_points_as_spheres=True, label="Cell Centroidss")

        return plot

    def fast_notebook_vis(self, plot, show_outline=True, show_centroids=True):

        # Set the grid dimensions: shape + 1 because we want to inject our values on the CELL data
        grid = pv.UniformGrid()
        grid.dimensions = np.array(self.shape) + 1
        # The bottom left corner of the data set
        grid.origin = self.minbound - self.unit * 0.5
        grid.spacing = self.unit  # These are the cell sizes along each axis
        # Add the data values to the cell data
        grid.cell_arrays["values"] = self.flatten(
            order="F").astype(float)  # Flatten the array!
        # filtering the voxels
        threshed = grid.threshold([0.9, 1.1])

        # adding the voxels: light red
        plot.add_mesh(threshed, color="#ff8fa3", opacity=0.3)
        # plot.add_mesh(threshed, show_edges=True, color="#ff8fa3", opacity=0.3, label="Cells")

        if show_outline:
            # adding the boundingbox wireframe
            plot.add_mesh(grid.outline(), color="grey")
            # plot.add_mesh(grid.outline(), color="grey", label="Domain")

        if show_centroids:
            # adding the voxel centeroids: red
            plot.add_points(pv.PolyData(self.centroids), color='#ff244c')
            # plot.add_mesh(pv.PolyData(self.centroids), color='#ff244c', point_size=5, render_points_as_spheres=True, label="Cell Centroidss")

        return plot

    def boolean_marching_cubes(self):

        # construct the boolean_marching_cubes stencil
        mc_stencil = create_stencil("boolean_marching_cube", 1)

        # getting shifts by expanding the stencil in the Fortran Order
        shifts = mc_stencil.expand('F')

        # pad the volume with zero in every direction
        # TODO make this an option instead of default
        volume = np.pad(self, (1, 1), mode='constant', constant_values=(0, 0))

        # the id of voxels (0,1,2, ... n)
        volume_inds = np.arange(volume.size).reshape(volume.shape)

        # gattering all the replacements in the collumns
        replaced_columns = [
            np.roll(volume_inds, shift, np.arange(3)).ravel() for shift in shifts]

        # stacking the columns
        cell_corners = np.stack(replaced_columns, axis=-1)

        # converting volume value (TODO: this needs to become a method of its own)
        volume_flat = volume.ravel()
        volume_flat[volume_flat > 0.5] = 1
        volume_flat[volume_flat < 0.5] = 0

        # replace neighbours by their value in volume
        neighbor_values = volume_flat[cell_corners]

        # computing the cell tile id
        # the powers of 2 in an array
        legend = 2**np.arange(8)

        # multiply the corner with the power of two, sum them, and reshape to the original volume shape
        tile_id = np.sum(legend * neighbor_values,
                         axis=1).reshape(volume.shape)

        # drop the last column, row and page (since cube-grid is 1 less than the voxel grid in every dimension)
        # TODO consider that removing padding would eliminate the need for this line
        cube_grid = tile_id[:-1, :-1, :-1]

        # initializing the lattice
        cube_lattice = lattice([self.minbound, self.maxbound + self.unit],
                               unit=self.unit, dtype=np.uint8, buffer=cube_grid, default_value=False)

        # set the values that are bigger than 0 (transfering values)
        cube_lattice[cube_grid > 0] = cube_grid[cube_grid > 0]

        return cube_lattice

    def find_connectivity(self, stencil):
        raise NotImplementedError


class cloud(np.ndarray):

    def __new__(subtype, point_array, dtype=float, buffer=None, offset=0,
                strides=None, order=None):

        # extracting the shape from point_array
        shape = point_array.shape
        # using the point_array as the buffer
        buffer = point_array.flatten(order="C")

        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to cloud.__array_finalize__
        obj = super(cloud, subtype).__new__(subtype, shape, dtype,
                                            buffer, offset, strides,
                                            order)

        # set the  'bounds' attribute
        obj.bounds = np.array([obj.min(axis=0), obj.max(axis=0)])

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(cloud, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. cloud():
        #    obj is None
        #    (we're in the middle of the cloud.__new__
        #    constructor, and self.bounds will be set when we return to
        #    cloud.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(cloud):
        #    obj is arr
        #    (type(obj) can be cloud)
        # From new-from-template - e.g cloud[:3]
        #    type(obj) is cloud
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'bounds', because this
        # method sees all creation of default objects - with the
        # cloud.__new__ constructor, but also with
        # arr.view(cloud).
        self.bounds = getattr(obj, 'bounds', None)
        # We do not need to return anything

    @property
    def minbound(self):
        return self.bounds[0]

    @property
    def maxbound(self):
        return self.bounds[1]

    def regularize(self, unit, **kwargs):
        """[summary]

        Arguments:
            unit {[float or array of floats]} -- [the unit separation between cells of lattice]

        Keyword Arguments:
            closed {[Boolean]} -- [False by default. If the cell intervals are closed intervals or not.]

        Raises:
            ValueError: [unit needs to be either a float or an array of floats that has the same dimension of the points in the point cloud]

        Returns:
            [lattice] -- [a boolian latice representing the rasterization of point cloud]
        """
        ####################################################
        # INPUTS
        ####################################################
        unit = np.array(unit)
        if unit.size != 1 and unit.size != self.bounds.shape[1]:
            raise ValueError(
                'the length of unit array needs to be either 1 or equal to the dimension of point cloud')
        elif unit.size == 1:
            unit = np.tile(unit, (1, self.bounds.shape[1]))

        closed = kwargs.get('closed', False)

        ####################################################
        # PROCEDURE
        ####################################################

        if closed:
            # retrieve the identity matrix as a list of main axes
            axes = np.identity(unit.size).astype(int)
            # R3 to Z3 : finding the closest voxel to each point
            point_scaled = self / unit
            # shift the hit points in each 2-dimension (n in 1-axes) backward and formard (s in [-1,1]) and rint all the possibilities
            vox_ind = [np.rint(point_scaled + unit * n * s * 0.001)
                       for n in (1-axes) for s in [-1, 1]]
            print([unit * n * s for n in (1-axes) for s in [-1, 1]])
            vox_ind = np.vstack(vox_ind)
            print("close")
        else:
            vox_ind = np.rint(self / unit).astype(int)
        print("here")
        # removing repetitions
        unique_vox_ind = np.unique(vox_ind, axis=0).astype(int)

        # mapping the voxel indices to real space
        reg_pnt = unique_vox_ind * unit

        # initializing the volume
        l = lattice([self.minbound, self.maxbound], unit=unit,
                    dtype=bool, default_value=False)

        # map the indices to start from zero
        mapped_ind = unique_vox_ind - np.rint(l.bounds[0]/l.unit).astype(int)

        # setting the occupied voxels to True
        l[mapped_ind[:, 0], mapped_ind[:, 1], mapped_ind[:, 2]] = True

        ####################################################
        # OUTPUTS
        ####################################################

        return l

    def fast_vis(self, plot, color='#2499ff'):

        # adding the original point cloud: blue
        plot.add_mesh(pv.PolyData(self), color=color, point_size=3,
                      render_points_as_spheres=True, label="Point Cloud")

        return plot

    def fast_notebook_vis(self, plot, color='#2499ff'):

        # adding the original point cloud: blue
        plot.add_points(pv.PolyData(self), color=color)

        return plot


class stencil(np.ndarray):

    def __new__(subtype, point_array, ntype="Custom", origin=np.array([0, 0, 0]), dtype=int, buffer=None, offset=0,
                strides=None, order=None):

        # extracting the shape from point_array
        shape = point_array.shape
        # using the point_array as the buffer
        buffer = point_array.flatten(order="C")

        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to stencil.__array_finalize__
        obj = super(stencil, subtype).__new__(subtype, shape, dtype,
                                              buffer, offset, strides,
                                              order)

        # set the neighbourhood type
        obj.ntype = ntype
        # set the origin
        obj.origin = origin
        # set the  'bounds' attribute
        shape_arr = np.array(shape)
        obj.bounds = np.array([shape_arr * 0, shape_arr - 1]) - origin

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(stencil, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. stencil():
        #    obj is None
        #    (we're in the middle of the stencil.__new__
        #    constructor, and self.bounds will be set when we return to
        #    stencil.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(stencil):
        #    obj is arr
        #    (type(obj) can be stencil)
        # From new-from-template - e.g stencil[:3]
        #    type(obj) is stencil
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'properties', because this
        # method sees all creation of default objects - with the
        # stencil.__new__ constructor, but also with
        # arr.view(stencil).
        self.bounds = getattr(obj, 'bounds', None)
        self.ntype = getattr(obj, 'ntype', None)
        self.origin = getattr(obj, 'origin', None)

        # We do not need to return anything

    def __array_wrap__(self, array, context=None):

        temp = np.array(array)

        # checking if the array has any value other than 0, and 1
        np.place(temp, temp > 0.5, [1])
        np.place(temp, temp < 0.5, [0])

        return stencil(temp, ntype="custom", origin=self.origin)

    @property
    def minbound(self):
        return self.bounds[0]

    @property
    def maxbound(self):
        return self.bounds[1]

    def expand(self, sort="dist"):
        # list the locations
        locations = self.origin - np.argwhere(self)

        # check the sorting method

        if sort == "dist":  # Sorted Based on the distance from origin
            # calculating the distance of each neighbour
            sums = np.abs(locations).sum(axis=1)
            # sorting to identify the main cell
            order = np.argsort(sums)

        elif sort == "F":  # Fortran Sort, used for Boolean Marching Cubes
            order = np.arange(self.size).reshape(self.shape).flatten('F')

        # sort and return
        return locations[order].astype(int)

    def set_index(self, index, value):
        ind = np.array(index) + self.origin
        if ind.size != 3:
            raise ValueError(" the index needs to have three components")
        self[ind[0], ind[1], ind[2]] = value


def create_stencil(type_str, steps, clip=None):
    # check if clip is specified. if it is not, set it to the steps
    if clip == None:
        clip = steps
    # von neumann neighborhood
    if type_str == "von_neumann":
        # https://en.wikipedia.org/wiki/Von_Neumann_neighborhood

        # claculating all the possible shifts to apply to the array
        shifts = np.array(list(itertools.product(
            list(range(-clip, clip+1)), repeat=3)))

        # the number of steps that the neighbour is appart from the cell (setp=1 : 6 neighbour, step=2 : 18 neighbours, step=3 : 26 neighbours)
        shift_steps = np.sum(np.absolute(shifts), axis=1)

        # check the number of steps
        chosen_shift_ind = np.argwhere(shift_steps <= steps).ravel()

        # select the valid indices from shifts variable, transpose them to get separate indicies in rows, add the number of steps to make this an index
        locs = np.transpose(shifts[chosen_shift_ind]) + clip

        # inilize the stencil
        s = np.zeros((clip*2+1, clip*2+1, clip*2+1)).astype(int)

        # fill in the stencil
        s[locs[0], locs[1], locs[2]] = 1

        return stencil(s, ntype=type_str, origin=np.array([clip, clip, clip]))

    elif type_str == "moore":
        # https://en.wikipedia.org/wiki/Moore_neighborhood

        # claculating all the possible shifts to apply to the array
        shifts = np.array(list(itertools.product(
            list(range(-clip, clip+1)), repeat=3)))

        # the number of steps that the neighbour is appart from the origin cell
        shift_steps = np.max(np.absolute(shifts), axis=1)

        # check the number of steps
        chosen_shift_ind = np.argwhere(shift_steps <= steps).ravel()

        # select the valid indices from shifts variable, transpose them to get separate indicies in rows, add the number of steps to make this an index
        locs = np.transpose(shifts[chosen_shift_ind]) + clip

        # inilize the stencil
        s = np.zeros((clip*2+1, clip*2+1, clip*2+1)).astype(int)

        # fill in the stencil
        s[locs[0], locs[1], locs[2]] = 1

        return stencil(s, ntype=type_str, origin=np.array([clip, clip, clip]))

    elif type_str == "boolean_marching_cube":

        # # shifts to check 8 corner of cube (multiply by -1 since shift goes backward)
        # shifts = np.array([
        #     [0, 0, 0],  # 1
        #     [1, 0, 0],  # 2
        #     [0, 1, 0],  # 4
        #     [1, 1, 0],  # 8
        #     [0, 0, 1],  # 16
        #     [1, 0, 1],  # 32
        #     [0, 1, 1],  # 64
        #     [1, 1, 1]   # 128
        # ])*-1

        # # the number of steps that the neighbour is appart from the origin cell
        # shift_steps = np.max(np.absolute(shifts), axis=1)

        # # check the number of steps
        # chosen_shift_ind = np.argwhere(shift_steps <= steps).ravel()

        # # select the valid indices from shifts variable, transpose them to get separate indicies in rows, add the number of steps to make this an index
        # locs = np.transpose(shifts[chosen_shift_ind]) + clip

        # inilize the stencil
        s = np.ones((2, 2, 2)).astype(int)

        # # fill in the stencil
        # s[locs[0], locs[1], locs[2]] = 1

        return stencil(s, ntype=type_str, origin=np.array([0, 0, 0]))

    else:
        raise ValueError(
            'non-valid neighborhood type for stencil creation')


def scatter(bounds, count):
    """[summary]

    Arguments:
        bounds {[2d array]} -- [array of two vectors, indicating the bounding box of the scattering envelope with a minimum and maximum of the bounding box]
        count {[int]} -- [number of the points to scatter within the bounding box]

    Returns:
        [cloud] -- [returns a cloud object countaing the coordinates of the scattered points]
    """
    point_array = np.random.uniform(
        bounds[0], bounds[1], (count, bounds.shape[1]))
    return cloud(point_array)


def cloud_from_csv(file_path, delimiter=','):

    point_array = np.genfromtxt(file_path, delimiter=delimiter)
    return cloud(point_array)


def lattice_from_csv(file_path):

    # read the voxel 3-dimensional indices
    ind_flat = np.genfromtxt(file_path, delimiter=',',
                             skip_header=8, usecols=(0, 1, 2)).astype(int)

    # read the voxel values
    vol_flat = np.genfromtxt(
        file_path, delimiter=',', skip_header=8, usecols=(3)).astype(int)

    # read volume meta data
    meta = np.genfromtxt(
        file_path, delimiter='-', skip_header=1, max_rows=3, usecols=(1, 2, 3))
    unit = meta[0]
    min_bound = meta[1]
    volume_shape = meta[2].astype(int)
    max_bound = min_bound + unit * volume_shape

    # reshape the 1d array to get 3d array
    vol = vol_flat.reshape(volume_shape)

    # initializing the lattice
    l = lattice([min_bound, max_bound], unit=unit,
                dtype=bool, default_value=False)

    # setting the latice equal to volume
    l[ind_flat[:, 0], ind_flat[:, 1], ind_flat[:, 2]
      ] = vol[ind_flat[:, 0], ind_flat[:, 1], ind_flat[:, 2]]

    return l


def find_neighbours(lattice, stencil):

    # flatten the lattice
    lattice_flat = lattice.ravel()

    # the id of voxels (0,1,2, ... n)
    lattice_inds = np.arange(lattice.size).reshape(lattice.shape)

    # removing the indecies that are not filled in the volume
    lattice_inds = ((lattice_inds + 1) * lattice) - 1

    # offset the 1-dimensional indices of the voxels that is rshaped to volume shape with value -1
    lattice_inds_paded = np.pad(lattice_inds, (1, 1), mode='constant',
                                constant_values=(-1, -1))

    # flatten
    lattice_inds_paded_flat = lattice_inds_paded.ravel()

    # index of padded cells in flatten
    origin_flat_ind = np.argwhere(lattice_inds_paded_flat != -1).ravel()

    # retrievig all the possible shifts corresponding to the neighbours defined in stencil
    shifts = expand_stencil(stencil)

    # gattering all the replacements in the collumns
    replaced_columns = [
        np.roll(lattice_inds_paded, shift, np.arange(3)).ravel() for shift in shifts]

    # stacking the columns and removing the pads (and also removing the neighbours of the empty voxels since we have tagged them -1 like paddings)
    cell_neighbors = np.stack(replaced_columns, axis=-1)[origin_flat_ind]

    return cell_neighbors


def marching_cube_vis(p, cube_lattice, style_str):

    # extract cube indicies
    cube_ind = np.transpose(np.indices(cube_lattice.shape),
                            (1, 2, 3, 0)).reshape(-1, 3)
    # extract cube positions
    cube_pos = (cube_ind - 0.5) * cube_lattice.unit + cube_lattice.minbound

    # extract cube tid
    cube_tid = cube_lattice.ravel()

    # remove the cube position and tid where tid is 0
    filled_cube_pos = cube_pos[cube_tid > 0]
    filled_cube_tid = cube_tid[cube_tid > 0]

    if style_str != "chamfer":
        raise ValueError(
            "Meshing style is not valid. Valid styles are: ['chamfer']")

    # load tiles
    tiles = [0]
    for i in range(1, 256):
        tile_path = os.path.join(
            file_directory, "resources/mc_tiles", style_str, "Tile_{0:03d}.obj".format(i))
        tile = pv.read(tile_path)
        tile.points *= cube_lattice.unit
        tiles.append(tile)

    new_points = tiles[filled_cube_tid[0]].points + filled_cube_pos[0]
    new_faces = tiles[filled_cube_tid[0]].faces.reshape(-1, 4)

    # merge tiles
    for i in range(1, filled_cube_tid.size):
        tile = tiles[filled_cube_tid[i]]
        # add the faces list, changing the point numbers
        new_faces = np.concatenate(
            (new_faces, tile.faces.reshape(-1, 4) + np.array([0, 1, 1, 1])*new_points.shape[0]), axis=0)
        # add the new points, change the position based on the location
        new_points = np.concatenate(
            (new_points, tile.points + filled_cube_pos[i]), axis=0)

    # construct the new mesh and add it to plot
    new_tile = pv.PolyData(new_points, new_faces)
    p.add_mesh(new_tile, color='#abd8ff')

    return p


def mesh_sampling(mesh, unit, tol=1e-06, **kwargs):
    """This algorithm samples a mesh based on unit size

    Args:
        geo_mesh ([COMPAS Mesh]): [description]
        unit ([numpy array]): [Unit represents the unit size in the sampling grid. It needs to be one float value or an array-like with three elements. In case that a scalar is given it will used for all three dimensions]
        tol ([type], optional): [description]. Defaults to 1e-06.

    Returns:
        [type]: [description]
    """
    ####################################################
    # INPUTS
    ####################################################

    unit = np.array(unit)
    if unit.size == 1:
        unit = np.array([unit, unit, unit])
    elif unit.size != 3:
        raise ValueError(
            "unit needs to have three elements representing the unit size for mesh sampling in three dimensions")
    dim_num = unit.size
    multi_core_process = kwargs.get('multi_core_process', False)
    return_ray_origin = kwargs.get('return_ray_origin', False)

    # compare voxel size and tolerance and warn if it is not enough
    if min(unit) * 1e-06 < tol:
        warnings.warn(
            "Warning! The tolerance for rasterization is not small enough, it may result in faulty results or failure of rasterization. Try decreasing the tolerance or scaling the geometry.")

    ####################################################
    # Initialize the volumetric array
    ####################################################
    mesh_vertices, mesh_faces = mesh

    # retrieve the bounding box information
    mesh_bb_min = np.amin(mesh_vertices, axis=0)
    mesh_bb_max = np.amax(mesh_vertices, axis=0)
    mesh_bb_size = mesh_bb_max - mesh_bb_min

    # find the minimum index in discrete space
    mesh_bb_min_z3 = np.rint(mesh_bb_min / unit).astype(int)
    # calculate the size of voxelated volume
    vol_size = np.ceil((mesh_bb_size / unit)+1).astype(int)
    # initiate the 3d array of voxel space called volume
    vol = np.zeros(vol_size)

    ####################################################
    # claculate the origin and direction of rays
    ####################################################

    # increasing the vol_size by one to accomodate for shooting from corners
    vol_size_off = vol_size + 1
    # retriev the voxel index for ray origins
    hit_vol_ind = np.indices(vol_size_off)
    vol_ind_trans = np.transpose(hit_vol_ind) + mesh_bb_min_z3
    hit_vol_ind = np.transpose(vol_ind_trans)

    # retieve the ray origin indicies
    ray_orig_ind = [np.take(hit_vol_ind, 0, axis=d + 1).transpose((1,
                                                                   2, 0)).reshape(-1, 3) for d in range(dim_num)]
    ray_orig_ind = np.vstack(ray_orig_ind)

    # retrieve the direction of ray shooting for each origin point
    normals = np.identity(dim_num).astype(int)
    # tile(stamp) the X-ray direction with the (Y-direction * Z-direction) . Then repeat this for all dimensions
    ray_dir = [np.tile(normals[d], (vol_size_off[(d+1) % dim_num]*vol_size_off[(d+2) % dim_num], 1))
               for d in range(dim_num)]  # this line has a problem given the negative indicies are included now
    ray_dir = np.vstack(ray_dir)

    # project the ray origin + shift it with half of the voxel siz to move it to corners of the voxels
    ray_orig = ray_orig_ind * unit + unit * -0.5  # * (1 - ray_dir)

    # project the ray origin
    proj_ray_orig = ray_orig * (1 - ray_dir)

    ####################################################
    # intersection
    ####################################################

    samples = []

    # check if multiprocessing is allowed
    if multi_core_process:
        # open the context manager
        with concurrent.futures.ProcessPoolExecutor() as executor:
            # submit the processes
            # results = [executor.submit(intersect, geo_mesh, face, unit, mesh_bb_size, ray_orig, proj_ray_orig, ray_dir, tol) for face in geo_mesh.faces()]
            results = [executor.submit(intersect, mesh_vertices[face], unit, mesh_bb_size,
                                       ray_orig, proj_ray_orig, ray_dir, tol) for face in mesh_faces]
            # fetch the results
            for f in concurrent.futures.as_completed(results):
                samples.extend(f.result())
    else:
        # iterate over the faces
        # for face in geo_mesh.faces():
        for face in mesh_faces:
            # print(face)
            # print(type(face))
            # id = np.array(face)
            # print(mesh_vertices[id])
            # print(mesh_vertices[np.array(face).astype(int)])
            # print(mesh_vertices[np.array(face)])

            face_hit_pos = intersect(
                mesh_vertices[face], unit, mesh_bb_size, ray_orig, proj_ray_orig, ray_dir, tol)
            samples.extend(face_hit_pos)

    ####################################################
    # OUTPUTS
    ####################################################

    # set the output list
    out = [cloud(np.array(samples))]
    # check if the ray origins are requested
    if return_ray_origin:
        out.append(cloud(np.array(ray_orig)))

    # if the list has more than one item, return it as a tuple, if it has only one item, return the item itself
    return tuple(out) if len(out) > 1 else out[0]


def intersect(face_verticies_xyz, unit, mesh_bb_size, ray_orig, proj_ray_orig, ray_dir, tol):
    face_hit_pos = []

    # check if the face is a triangle
    if face_verticies_xyz.shape[0] != 3:
        return([])

    # check if any coordinate of the projected ray origin is in betwen the max and min of the coordinates of the face
    min_con = proj_ray_orig >= np.amin(
        face_verticies_xyz, axis=0)*(1 - ray_dir)
    max_con = proj_ray_orig <= np.amax(
        face_verticies_xyz, axis=0)*(1 - ray_dir)
    in_range_rays = np.all(min_con * max_con, axis=1)

    # retrieve the ray indicies that are in range
    in_rang_ind = np.argwhere(in_range_rays).flatten()

    # iterate over the rays
    for ray in in_rang_ind:
        # retrieve ray direction
        direction = ray_dir[ray]
        # retrieve ray origin
        orig_pos = ray_orig[ray]
        # calc the destination of ray (max distance that it needs to travel)
        # this line has a problem given the negative indicies are included now
        dest_pos = orig_pos + direction * mesh_bb_size

        # intersction
        # compas version
        # hit_pt = compas.geometry.intersection_line_triangle((orig_pos, dest_pos), face_verticies_xyz, tol=tol)
        # Translated from Pirouz C#
        hit_pt = TriangleLineIntersect(
            (orig_pos, dest_pos), face_verticies_xyz, tol=tol)
        if hit_pt is not None:
            face_hit_pos.append(hit_pt)

    return(face_hit_pos)


def TriangleLineIntersect(L, Vx, tol=1e-06):
    """
    Computing the intersection of a line with a triangle
    Algorithm from http://geomalgorithms.com/a06-_intersect-2.html
    C# implementation from https://github.com/Pirouz-Nourian/Topological_Voxelizer_CSharp/blob/master/Voxelizer_Functions.cs

    Args:
        L ([2d np array]): List of two points specified by their coordinates
        Vx ([2d np array]): List of three points specified by their coordinates
        tol ([float], optional): tolerance. Defaults to 1e-06.

    Raises:
        ValueError: If the triangle contains more than three vertices

    Returns:
        [np array]: [description]
    """

    ####################################################
    # INPUTS
    ####################################################

    if len(Vx) != 3:
        raise ValueError('A triangle needs to have three vertexes')

    ####################################################
    # PROCEDURE
    ####################################################

    # finding U & V vectors
    O = Vx[0]
    U = Vx[1] - Vx[0]
    V = Vx[2] - Vx[0]
    # finding normal vector
    N = surface_normal_newell_vectorized(Vx)  # np.cross(U, V)

    Nomin = np.dot((O - L[0]), N)
    Denom = np.dot(N, (L[1] - L[0]))

    if Denom != 0:
        alpha = Nomin / Denom

        # L[0] + np.dot(alpha, (L[1] - L[0])): parameter along the line where it intersects the plane in question, only if not paralell to the plane
        W = L[0] + np.dot(alpha, (L[1] - L[0])) - O

        UU = np.dot(U, U)
        VV = np.dot(V, V)
        UV = np.dot(U, V)
        WU = np.dot(W, U)
        WV = np.dot(W, V)

        STDenom = UV**2 - UU * VV

        s = (UV * WV - VV * WU) / STDenom
        t = (UV * WU - UU * WV) / STDenom

    ####################################################
    # OUTPUTS
    ####################################################

        if s + tol >= 0 and t + tol >= 0 and s + t <= 1 + 2*tol:
            Point = O + s * U + t * V
            return Point
        else:
            return None
    else:
        return None


def surface_normal_newell_vectorized(poly):
    """    
    https://stackoverflow.com/questions/39001642/calculating-surface-normal-in-python-using-newells-method
    Newell Method explained here: https://www.researchgate.net/publication/324921216_Topology_On_Topology_and_Topological_Data_Models_in_Geometric_Modeling_of_Space

    Args:
        poly ([2d np array]): List of vertices specified by their coordinates 

    Raises:
        ValueError: [description]

    Returns:
        [type]: [description]
    """

    # This section is the vectorized equivalent of this code

    """
    n = np.array([0.0, 0.0, 0.0])

    for i in range(3):
        j = (i+1) % 3
        n[0] += (poly[i][1] - poly[j][1]) * (poly[i][2] + poly[j][2])
        n[1] += (poly[i][2] - poly[j][2]) * (poly[i][0] + poly[j][0])
        n[2] += (poly[i][0] - poly[j][0]) * (poly[i][1] + poly[j][1])
    """

    poly_10 = np.roll(poly, [-1, 0], np.arange(2))
    poly_01 = np.roll(poly, [0, -1], np.arange(2))
    poly_11 = np.roll(poly, [-1, -1], np.arange(2))

    n = np.roll(np.sum((poly - poly_10) * (poly_01 + poly_11), axis=0), -1, 0)

    norm = np.linalg.norm(n)
    if norm == 0:
        raise ValueError('zero norm')
    else:
        normalised = n/norm

    return normalised


def load_mesh(mesh_path):
    # load the mesh using pyvista
    pv_mesh = pv.read(mesh_path)

    # check if all of the faces are triangles
    if not pv_mesh.is_all_triangles():
        raise ValueError('All faces need to be triangles!')

    # extract faces and vertices
    v = np.array(pv_mesh.points)
    f = pv_mesh.faces.reshape(-1, 4)[:, 1:]

    # return them as 3d numpy arrays
    return np.array(v).astype(np.float64), np.array(f).astype(np.int64)




"""
Algorithm Sketches that need to be integrated with the rest of library
"""

import itertools
import numpy as np
import pandas as pd
import networkx as nx
import compas
from compas.datastructures import Mesh
import concurrent.futures
import warnings
from topogenesis.data import find_neighbours

__author__ = "Shervin Azadi, and Pirouz Nourian"
__copyright__ = "???"
__credits__ = ["Shervin Azadi", "Pirouz Nourian"]
__license__ = "???"
__version__ = "0.0.2"
__maintainer__ = "Shervin Azadi"
__email__ = "shervinazadi93@gmail.com"
__status__ = "Dev"


def connectivity_graph(vol, steps):

    ####################################################
    # Finding the voxel neighbours for each voxel (generalized)
    ####################################################

    cell_neighbors = find_neighbours(vol, steps)

    ####################################################
    # Creating the graph
    ####################################################

    # the code in this section is a vectorized equivalent of:
    """
    edges = []
    for cell_neigh in cell_neighbors:
        cell = cell_neigh[0]
        for neigh in cell_neigh[1:]:
            if neigh != -1 and neigh > cell:
                edges.append((cell, neigh))
    """

    # removing the index of the cell itself
    cell_only_neighs = cell_neighbors[:, 1:]

    # tile the first column (cell id) with the size of the number of the neighs
    cell_only_neighs_rind = np.tile(
        cell_neighbors[:, 0].reshape(-1, 1), (1, cell_only_neighs.shape[1]))

    # flatten the neighs and row indices
    neighs_flat = cell_only_neighs.ravel()
    neighs_row_ind_flat = cell_only_neighs_rind.ravel()

    # find th index of the neighbours with real indices (everything except -1)
    real_neigh_ind = np.argwhere(neighs_flat != -1).ravel()

    # stack to list together and transpose them
    cell_edge_order_trans = np.stack(
        [neighs_row_ind_flat[real_neigh_ind], neighs_flat[real_neigh_ind]])
    cell_edge_order = np.transpose(cell_edge_order_trans)

    # create graph based on the edges
    G = nx.Graph()
    G.add_edges_from([tuple(edge) for edge in cell_edge_order.tolist()])

    return G


def pnts_to_csv(pnts, filepath, **kwargs):

    metadata = kwargs.get('metadata', None)
    metawrite = False if metadata is None else True
    # point position to panada dataframe
    pnts_df = pd.DataFrame(
        {
            'PX': pnts[:, 0],
            'PY': pnts[:, 1],
            'PZ': pnts[:, 2],
        })

    # save to csv
    with open(filepath, 'w') as df_out:
        if metawrite:
            df_out.write('metadata:\n')
            metadata.to_csv(df_out, index=False,
                            header=False, float_format='%g')
            df_out.write('\n')
        df_out.write('points:\n')
        pnts_df.to_csv(df_out, index=True, float_format='%g')


def vol_to_pandas(vol):
    # get the indices of the voxels
    vol_3d_ind = np.indices(vol.shape)

    # flatten except the last dimension
    vol_3d_ind_flat = vol_3d_ind.transpose(1, 2, 3, 0).reshape(-1, 3)

    # flatten the volume
    vol_flat = vol.ravel()

    # volume data to panda dataframe
    vol_df = pd.DataFrame(
        {'IX': vol_3d_ind_flat[:, 0],
            'IY': vol_3d_ind_flat[:, 1],
            'IZ': vol_3d_ind_flat[:, 2],
            'value': vol_flat,
         })
    return vol_df


def vol_to_csv(vol, filepath, **kwargs):

    metadata = kwargs.get('metadata', None)
    metawrite = False if metadata is None else True
    # volume to panda dataframe
    vol_df = vol_to_pandas(vol)
    # panada-dataframe to csv

    with open(filepath, 'w') as df_out:
        if metawrite:
            df_out.write('metadata:\n')
            metadata.to_csv(df_out, index=False,
                            header=False, float_format='%g')
            df_out.write('\n')
        df_out.write('volume:\n')
        vol_df.to_csv(df_out, index=False, float_format='%g')


def CellularAutomata(data):

    # convert string to json
    if isinstance(data, str):
        data = json.loads(data)

    # retrieve the position list
    voxel_positions = np.array(data['voxel_positions'])

    # construct the volume
    minbound = voxel_positions.min(axis=0)
    maxbound = voxel_positions.max(axis=0)
    volume = np.zeros(maxbound-minbound + 1).astype(int)

    # fill in the volume
    mapped_ind = voxel_positions - minbound
    volume[mapped_ind[:, 0], mapped_ind[:, 1], mapped_ind[:, 2]] = 1

    # pad the volume with zero in every direction
    volume = np.pad(volume, (1, 1), mode='constant', constant_values=(0, 0))

    # the id of voxels (0,1,2, ... n)
    volume_inds = np.arange(volume.size).reshape(volume.shape)

    # computing all the possible shifts to apply to the array
    shifts = np.array(list(itertools.product([0, -1, 1], repeat=3)))

    # gattering all the replacements in the columns
    replaced_columns = [
        np.roll(volume_inds, shift, np.arange(3)).ravel() for shift in shifts]

    # stacking the columns
    cell_neighbors = np.stack(replaced_columns, axis=-1)

    # replace neighbours by their value in volume
    volume_flat = volume.ravel()
    neighbor_values = volume_flat[cell_neighbors]

    # sum the neighbour values
    neighbor_sum = neighbor_values.sum(axis=1)

    # turn off if less than 2 neighbours on
    volume_flat *= (neighbor_sum >= 2)

    # turn off if more than 6 neighbours on
    volume_flat *= (neighbor_sum <= 6)

    # turn on if 3 neighbours are on
    volume_flat[(neighbor_sum >= 3) * (neighbor_sum <= 4)] = 1

    # on-cells 1D-index
    on_cells_1d_ind = np.argwhere(volume_flat == 1).ravel()

    # unravel the indices to 3D-index
    on_cells_3d_ind = np.stack(np.unravel_index(
        on_cells_1d_ind, volume.shape), axis=-1)

    # map back the indices (add initial minimum and subtract 1 for the padded layer)
    new_voxel_positions = on_cells_3d_ind + minbound - 1

    # package the results
    result = {
        'voxel_positions': new_voxel_positions.tolist(),
        'to_update': data['to_update'] + ['voxels'] if 'to_update' in data else ['voxels'],
    }

    # extend the data dictionary with the result dictionary
    data.update(result)

    return result


def RandomWalkingAgents(data):

    # convert string to json
    if isinstance(data, str):
        data = json.loads(data)

    # retrieve the position list
    voxel_positions = np.array(data['voxel_positions'])

    # construct the volume
    minbound = voxel_positions.min(axis=0)
    maxbound = voxel_positions.max(axis=0)
    volume = np.zeros(maxbound-minbound + 1).astype(int)

    # fill in the volume
    mapped_ind = voxel_positions - minbound
    volume[mapped_ind[:, 0], mapped_ind[:, 1], mapped_ind[:, 2]] = 1

    # pad the volume with zero in every direction
    volume = np.pad(volume, (1, 1), mode='constant', constant_values=(0, 0))

    # the id of voxels (0,1,2, ... n)
    volume_inds = np.arange(volume.size).reshape(volume.shape)

    # shifts to check: self + 6 neighbours of each voxel
    shifts = np.array([
        [0, 0, 0],  # self
        [1, 0, 0],  # left
        [-1, 0, 0],  # right
        [0, 1, 0],  # up
        [0, -1, 0],  # down
        [0, 0, 1],  # back
        [0, 0, -1],  # front
    ])

    # gattering all the replacements in the columns
    replaced_columns = [
        np.roll(volume_inds, shift, np.arange(3)).ravel() for shift in shifts]

    # stacking the columns
    cell_neighbors = np.stack(replaced_columns, axis=-1)

    # replace neighbours by their value in volume and flip it: 0:occupied, 1:empty
    volume_flat = volume.ravel()
    neighbor_values_flipped = 1-volume_flat[cell_neighbors]

    # multiply the cell neighbours by their flipped values to remove all occupied neighbours. +1 and -1 is to prevent mixing the empty cells and the cell id 0, so we mark the empty cells -1 instead of 0
    empty_neighbours = neighbor_values_flipped * (cell_neighbors + 1) - 1

    # extracting the id and flipped value of neighbours of the filled voxels: current position of agents
    agent_neighbour_values_flipped = neighbor_values_flipped[np.where(
        volume_flat)]
    agent_neighbour_id = cell_neighbors[np.where(volume_flat)]

    # assigning random value to each neighbour (this can later be specified by a field instead of randomvalues)
    rand_shape = agent_neighbour_values_flipped.shape
    agn_neigh_priority = agent_neighbour_values_flipped * \
        np.random.rand(rand_shape[0], rand_shape[1])

    # getting the argmax to find the selected neighbours
    neigh_selection = agn_neigh_priority.argmax(axis=1)

    # extracting the voxel id of the current position and the next selected position of all agents
    cur_pos_id = agent_neighbour_id[:, 0]
    sel_pos_id = agent_neighbour_id[np.arange(
        neigh_selection.size), neigh_selection]

    # checking for unique ids in the new_position_id: to prevent two agents merging together by moving into one voxel
    __, unq_indices = np.unique(sel_pos_id, return_index=True)

    # setting the unique voxels in the selected position to current position: updating the cur_pos_id. (this is computationally cheaper than setting it in a new array)
    cur_pos_id[unq_indices] = sel_pos_id[unq_indices]

    # emptying the volume and filling the selected neighbours
    volume_flat *= 0
    volume_flat[cur_pos_id] = 1

    # on-cells 1D-index
    on_cells_1d_ind = np.argwhere(volume_flat == 1).ravel()

    # unravel the indices to 3D-index
    on_cells_3d_ind = np.stack(np.unravel_index(
        on_cells_1d_ind, volume.shape), axis=-1)

    # map back the indices (add initial minimum and subtract 1 for the padded layer)
    new_voxel_positions = on_cells_3d_ind + minbound - 1

    ###########################################################################
    # ATTENTION:
    # Agents are not able to merge anymore. if two agent has chose one single
    # voxel as the next position. one of would randomly be selected to move to
    # the next position and the other one will remain in his old position
    ###########################################################################

    # package the results
    result = {
        # 'to_update': ["hi there"],
        'voxel_positions': new_voxel_positions.tolist(),
        'to_update': data['to_update'] + ['voxels'] if 'to_update' in data else ['voxels'],
    }

    # extend the data dictionary with the result dictionary
    data.update(result)

    return result


def gradient(sfield):

    # partial derivatives
    dx = (sfield[:-2, 1:-1, 1:-1] - sfield[2:, 1:-1, 1:-1]) * 0.5
    dy = (sfield[1:-1, :-2, 1:-1] - sfield[1:-1, 2:, 1:-1]) * 0.5
    dz = (sfield[1:-1, 1:-1, :-2] - sfield[1:-1, 1:-1, 2:]) * 0.5

    # stack gradient
    g = np.stack([dx, dy, dz], axis=0)

    return(g)


def divergence(vfield):

    X = vfield[0]
    Y = vfield[0]
    Z = vfield[0]

    # partial derivatives
    dx = (X[:-2, 1:-1, 1:-1] - X[2:, 1:-1, 1:-1]) * 0.5
    dy = (Y[1:-1, :-2, 1:-1] - Y[1:-1, 2:, 1:-1]) * 0.5
    dz = (Z[1:-1, 1:-1, :-2] - Z[1:-1, 1:-1, 2:]) * 0.5

    # sum divergence
    d = dx + dy + dz

    return(d)




"""
topoGenesis DataStructure
"""
from scipy.spatial.transform import Rotation as sp_rotation
import numpy as np
import pandas as pd
import pyvista as pv
import itertools
import warnings
import os


file_directory = os.path.dirname(os.path.abspath(__file__))


class lattice(np.ndarray):
    """Lattice is subclass of NumPy ndarrays that is adapted to represent 
    a numerical field within a discrete 3dimensional space. It adds spatial 
    properties and functionalities to ndimensional arrays such as bounds, 
    unit, neighbourhood assessment and more.
    """

    def __new__(subtype, bounds, unit=1, dtype=float, buffer=None, offset=0,
                strides=None, order=None, default_value=None, orient=np.array([0., 0., 0., 1.])):

        # TODO: Add documentation for orient
        # extracting min and max from bound and discrtizing it
        bounds = np.array(bounds)
        minbound = np.rint(bounds[0] / unit).astype(int)
        maxbound = np.rint(bounds[1] / unit).astype(int)
        bounds = np.array([minbound, maxbound])*unit

        # unit np array
        unit = np.array(unit)

        # if the unit vector size is 1, tile it with the size of minimum vector
        if unit.size == 1:
            unit = np.tile(unit, minbound.size)
        # raise value error if the size of unit is neighter 1 nor the length of the minimum
        elif unit.size != minbound.size:
            raise ValueError(
                'the length of unit array needs to be either 1 or equal to the min/max arrays')

        # calculating shape based on bounds and unit
        shape = 1 + maxbound - minbound

        # set default value
        if default_value != None:
            buffer = np.tile(default_value, shape.flatten())

        # Create the ndarray instance of our type
        obj = super(lattice, subtype).__new__(subtype, shape, dtype,
                                              buffer, offset, strides,
                                              order)

        # set the  'bounds' attribute
        obj.bounds = bounds
        # set the attribute 'unit' to itself
        obj.unit = unit
        # set the 'orient' attribute
        obj.orient = np.array(orient)

        # init an empty connectivity
        obj.connectivity = None
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):

        if obj is None:
            return

        # adding the attributes to self
        self.bounds = getattr(obj, 'bounds', None)
        self.bounds = getattr(obj, 'bounds', None)
        self.unit = getattr(obj, 'unit', None)
        self.orient = getattr(obj, 'orient', None)
        self.connectivity = getattr(obj, 'connectivity', None)
        # TODO need to add the origin atribute

    @property
    def minbound(self):
        """Real minimum bound of the lattice

        Returns:
            numpy.ndarray: real minimum bound
        """
        return self.bounds[0]

    @property
    def maxbound(self):
        """Real maximum bound of the lattice

        Returns:
            numpy.ndarray: real maximum bound
        """
        return self.bounds[1]

    @property
    def centroids(self, threshold=0.0):
        """Extracts the centroid of cells that have a positive or True value and returns them as a point cloud

        Args:
            threshold (float, optional): exclusive lower bound of the values to assume existence of the voxel. Defaults to 0.0.

        Returns:
            topogenesis.Cloud: a point cloud representing the centroids of the lattice cells
        """

        # extract the indices of the True values # with sparse matrix we don't need to search
        point_array = np.argwhere(self > threshold)
        # convert to float
        point_array = point_array.astype(float)
        # scale by unit
        point_array *= self.unit
        # translate by minimum
        point_array += self.minbound
        # orient the points
        if np.sum(np.abs(self.orient[:3])) > 0.001 or self.orient[3] != 1: # check if the orientation is required
            r = sp_rotation.from_quat(self.orient)
            point_array = r.apply(point_array)
        # return as a point cloud
        return cloud(point_array, dtype=float)

    def centroids_threshold(self, threshold=0.0):
        """Extracts the centroid of cells that have a positive or True value and returns them as a point cloud

        Args:
            threshold (float, optional): exclusive lower bound of the values to assume existence of the voxel. Defaults to 0.0.

        Returns:
            topogenesis.Cloud: a point cloud representing the centroids of the lattice cells
        """

        # extract the indices of the True values # with sparse matrix we don't need to search
        point_array = np.argwhere(self > threshold)
        # convert to float
        point_array = point_array.astype(float)
        # scale by unit
        point_array *= self.unit
        # translate by minimum
        point_array += self.minbound
        # orient the points
        if np.sum(np.abs(self.orient[:3])) > 0.001 or self.orient[3] != 1: # check if the orientation is required
            r = sp_rotation.from_quat(self.orient)
            point_array = r.apply(point_array)
        # return as a point cloud
        return cloud(point_array, dtype=float)

    @property
    def indices(self):
        """Creates one-dimensional integer indices for cells in the lattice

        Returns:
            topogenesis.Lattice: integer lattice of indices 
        """

        ind = np.arange(self.size).reshape(self.shape)
        return to_lattice(ind.astype(int), self)

    def fast_vis(self, plot, show_outline: bool = True, show_centroids: bool = True, color = "#ff8fa3", opacity=0.3):
        """Adds the basic lattice features to a pyvista plotter and returns it. 
        It is mainly used to rapidly visualize the content of the lattice 
        for visual confirmation

        Args:
            plot (pyvista.Plotter): a pyvista plotter
            show_outline (bool, optional): If `True`, adds the bounding box of the lattice to the plot
            show_centroids (bool, optional): If `True`, adds the centroid of cells to the plot

        Returns:
            pyvista.Plotter: the same pyvista plotter containing lattice features

        ** Usage Example: **
        ```python
        p = pyvista.Plotter()
        lattice.fast_vis(p)
        ```
        """
        # TODO: Add documentation for color and opacity

        # Set the grid dimensions: shape + 1 because we want to inject our values on the CELL data
        grid = pv.UniformGrid()
        grid.dimensions = np.array(self.shape) + 1
        # The bottom left corner of the data set
        grid.origin = self.minbound - self.unit * 0.5
        grid.spacing = self.unit  # These are the cell sizes along each axis
        # Add the data values to the cell data
        grid.cell_arrays["values"] = self.flatten(
            order="F").astype(float)  # Flatten the array!
        # filtering the voxels
        threshed = grid.threshold([0.9, 1.1])

        # applying the orientation of the lattice
        if np.sum(np.abs(self.orient[:3])) > 0.001 or self.orient[3] != 1: # check if the orientation is required
            Rz = sp_rotation.from_quat(self.orient)
            threshed.points = Rz.apply(threshed.points)

        # adding the voxels: light red
        plot.add_mesh(threshed, show_edges=True, color=color,
                      opacity=opacity, label="Cells")

        if show_outline:
            # adding the boundingbox wireframe
            wireframe = grid.outline()
            Rz = sp_rotation.from_quat(self.orient)
            wireframe.points = Rz.apply(wireframe.points)
            plot.add_mesh(wireframe, color="grey", label="Domain")

        if show_centroids:
            # adding the voxel centroids: red
            plot.add_mesh(pv.PolyData(self.centroids), color='#ff244c', point_size=5,
                          render_points_as_spheres=True, label="Cell Centroids")

        return plot

    def fast_volumetric_vis(self, plot, show_outline: bool = True, show_centroids: bool = True, cmap="coolwarm", clim=[0.5, 1.0], opacity=np.array([0,0.9,0.9,0.9,0.9,0.9,0.9]), value_tag="Value"):
        # TODO: resolve the orientation of the volumetric visualization
        # TODO: Add doc strings

        # Create the spatial reference
        grid = pv.UniformGrid()

        # Set the grid dimensions: shape because we want to inject our values
        grid.dimensions = self.shape
        # The bottom left corner of the data set
        grid.origin = self.minbound
        # These are the cell sizes along each axis
        grid.spacing = self.unit

        # Add the data values to the cell data
        grid.point_arrays[value_tag] = self.flatten(order="F")  # Flatten the Lattice

            
        # adding the volume
        plot.add_volume(grid, cmap=cmap, clim=clim,opacity=opacity, shade=True)

        # # Set the grid dimensions: shape + 1 because we want to inject our values on the CELL data
        # grid = pv.UniformGrid()
        # grid.dimensions = np.array(self.shape) + 1
        # # The bottom left corner of the data set
        # grid.origin = self.minbound - self.unit * 0.5
        # grid.spacing = self.unit  # These are the cell sizes along each axis
        # # Add the data values to the cell data
        # grid.cell_arrays["values"] = self.flatten(
        #     order="F").astype(float)  # Flatten the array!
        # # filtering the voxels
        # threshed = grid.threshold([0.9, 1.1])

        # # applying the orientation of the lattice
        # if np.sum(np.abs(self.orient[:3])) > 0.001 or self.orient[3] != 1: # check if the orientation is required
        #     Rz = sp_rotation.from_quat(self.orient)
        #     threshed.points = Rz.apply(threshed.points)

        # # adding the voxels: light red
        # plot.add_mesh(threshed, show_edges=True, color=color,
        #               opacity=opacity, label="Cells")

        # if show_outline:
        #     # adding the boundingbox wireframe
        #     wireframe = grid.outline()
        #     Rz = sp_rotation.from_quat(self.orient)
        #     wireframe.points = Rz.apply(wireframe.points)
        #     plot.add_mesh(wireframe, color="grey", label="Domain")

        # if show_centroids:
        #     # adding the voxel centroids: red
        #     plot.add_mesh(pv.PolyData(self.centroids), color='#ff244c', point_size=5,
        #                   render_points_as_spheres=True, label="Cell Centroids")

        return plot

    def fast_notebook_vis(self, plot, show_outline: bool = True, show_centroids: bool = True):
        """Adds the basic lattice features to a pyvista ITK plotter and returns it. 
        ITK plotters are specifically used in notebooks to plot the geometry inside 
        the notebook environment It is mainly used to rapidly visualize the content 
        of the lattice for visual confirmation

        Args:
            plot (pyvista.PlotterITK): a pyvista ITK plotter
            show_outline (bool, optional): If `True`, adds the bounding box of the lattice to the plot
            show_centroids (bool, optional): If `True`, adds the centroid of cells to the plot

        Returns:
            pyvista.PlotterITK: pyvista ITK plotter containing lattice features such as cells, bounding box, cell centroids

        ** Usage Example: **
        ```python
        p = pyvista.PlotterITK()
        lattice.fast_notebook_vis(p)    
        ```
        """

        # Set the grid dimensions: shape + 1 because we want to inject our values on the CELL data
        grid = pv.UniformGrid()
        grid.dimensions = np.array(self.shape) + 1
        # The bottom left corner of the data set
        grid.origin = self.minbound - self.unit * 0.5
        grid.spacing = self.unit  # These are the cell sizes along each axis
        # Add the data values to the cell data
        grid.cell_arrays["values"] = self.flatten(
            order="F").astype(float)  # Flatten the array!
        # filtering the voxels
        threshed = grid.threshold([0.9, 1.1])

        # adding the voxels: light red
        plot.add_mesh(threshed, color="#ff8fa3", opacity=0.3)
        # plot.add_mesh(threshed, show_edges=True, color="#ff8fa3", opacity=0.3, label="Cells")

        if show_outline:
            # adding the boundingbox wireframe
            plot.add_mesh(grid.outline(), color="grey")
            # plot.add_mesh(grid.outline(), color="grey", label="Domain")

        if show_centroids:
            # adding the voxel centroids: red
            plot.add_points(pv.PolyData(self.centroids), color='#ff244c')
            # plot.add_mesh(pv.PolyData(self.centroids), color='#ff244c', point_size=5, render_points_as_spheres=True, label="Cell Centroids")

        return plot

    def boolean_marching_cubes(self):
        """This is a polygonization method. It converts the lattice to a boolean lattice and runs a boolean marching cube on the lattice. 

        Returns:
            topogenesis.Lattice: an integer lattice that contains the tile-id at each cell
        """

        # construct the boolean_marching_cubes stencil
        mc_stencil = create_stencil("boolean_marching_cube", 1)

        # retrieve the value of the neighbours
        cell_corners = self.find_neighbours(mc_stencil, order="F")

        # converting volume value (TODO: this needs to become a method of its own)
        volume_flat = self.ravel()
        volume_flat[volume_flat > 0.0] = 1
        volume_flat[volume_flat <= 0.0] = 0

        # replace neighbours by their value in volume
        neighbor_values = volume_flat[cell_corners]

        # computing the cell tile id
        # the powers of 2 in an array
        legend = 2**np.arange(8)

        # multiply the corner with the power of two, sum them, and reshape to the original volume shape
        tile_id = np.sum(legend * neighbor_values,
                         axis=1).reshape(self.shape)

        # drop the last column, row and page (since cube-grid is 1 less than the voxel grid in every dimension)
        # TODO consider that by implementing the origin attribute in lattice this may have to change
        cube_grid = tile_id[:-1, :-1, :-1]

        # convert the array to lattice
        cube_lattice = to_lattice(
            cube_grid, minbound=self.minbound, unit=self.unit)

        return cube_lattice

    def find_connectivity(self, stencil):
        raise NotImplementedError

    def to_csv(self, filepath: str):
        """This method saves the lattice to a csv file

        Args:
            filepath: path to the csv file
        """
        # volume to panda dataframe
        vol_df = self.to_pandas()

        # specifying metadata and transposing it
        metadata = pd.DataFrame({
            'minbound': self.minbound,
            'shape': np.array(self.shape),
            'unit': self.unit,
        })

        with open(filepath, 'w') as df_out:

            metadata.to_csv(df_out, index=False,
                            header=True, float_format='%g', line_terminator='\n')

            df_out.write('\n')

            vol_df.to_csv(df_out,
                          index=False,
                          float_format='%g', line_terminator='\n')

    def to_pandas(self):
        """This methods returns a pandas dataframe containing the lattice information with integer indices and value of the cell as columns.

        Returns:
            pandas.Dataframe: lattice represented in pandas dataframe
        """
        # get the indices of the voxels
        vol_3d_ind = np.indices(self.shape)

        # flatten except the last dimension
        vol_3d_ind_flat = vol_3d_ind.transpose(1, 2, 3, 0).reshape(-1, 3)

        # flatten the volume
        vol_flat = self.ravel()

        # volume data to panda dataframe
        vol_df = pd.DataFrame(
            {
                'IX': vol_3d_ind_flat[:, 0],
                'IY': vol_3d_ind_flat[:, 1],
                'IZ': vol_3d_ind_flat[:, 2],
                'value': vol_flat,
            })
        return vol_df

    # TODO change the default padding value to np.nan. current problem is with datatypes other than float
    def apply_stencil(self, stencil, border_condition: str = "pad_outside", padding_value: int = 0):
        """This method applies the function of a given stencil on the lattice and returns the result in a new lattice.

        Args:
            stencil (topogenesis.Stencil): 
                the stencil to be applied on the lattice
            border_condition (str, optional): 
                specifies how the border condition should be treated. The options are {"pad_outside", "pad_inside", "roll"}. "pad_outside" will offset the lattice in every direction by one step, and fill the new cells with the given `padding_value` and procedes to performing the computation; the resultant lattice in this case has the same shape as the initial lattice. "pad_inside" will perform the computation on the lattice, offsets inside by one cell from each side and returns the remainder cells; the resultant lattice is 2 cell smaller in each dimension than the original lattice. "roll" will assume that the end of each dimension is connected to the beginning of it and interprets the connectivity of the lattice with a rolling approach; the resultant lattice has the same shape is the original lattice. defaults to "pad_outside".
            padding_value (int, optional): 
                value used for padding in case the `border_condition` is set to "pad_outside".

        Raises:
            NotImplementedError: "pad_inside" is not implemented yet

        Returns:
            topogenesis.Lattice: a new lattice containing the result of the application of the stencil
        """

        if border_condition == "pad_outside":
            # pad the volume with zero in every direction
            padded_arr = np.pad(self, (1, 1),
                                mode='constant',
                                constant_values=(padding_value, padding_value))
            # convert to lattice
            self_padded = to_lattice(padded_arr,
                                     self.minbound - self.unit,
                                     unit=self.unit)

        elif border_condition == "pad_inside":
            raise NotImplementedError

        elif border_condition == "roll":
            self_padded = to_lattice(np.copy(self), self)

        # find the neighbours based on the stencil
        cell_neighbors = self_padded.find_neighbours(stencil)

        # replace neighbours by their value in volume
        self_padded_flat = self_padded.ravel()
        neighbor_values = self_padded_flat[cell_neighbors]

        # apply the function to the neighbour values
        applied = stencil.function(neighbor_values, axis=1)

        # reshape the neighbour applied into the original lattice shape
        applied_3d = applied.reshape(self_padded.shape)

        # reverse the padding procedure
        if border_condition == "pad_outside":
            # trim the padded dimensions
            applied_3d_trimmed = applied_3d[1:-1, 1:-1, 1:-1]

        elif border_condition == "pad_inside":
            raise NotImplementedError

        elif border_condition == "roll":
            applied_3d_trimmed = applied_3d

        return to_lattice(applied_3d_trimmed, self)

    def arg_apply_stencil(self, arg_lattice, stencil, border_condition: str = "pad_outside", padding_value: int = 0):
        """Applies the function (should be argument function) of the stencil on the original lattice, extracts the value of the same cell of the argument lattice, fills a new lattice and returns it. If the argument lattice contains one-dimensional ordering of the original lattice, this would function as an argument function. 

        Args:
            arg_lattice (topogenesis.Lattice): 
                the argument lattice. The values in this lattice will be extracted by the argument function and used to fill the new lattice
            stencil (topogenesis.Stencil): 
                the stencil to be applied on the lattice. This stencil should contain an "argument function"
            border_condition (str, optional): 
                specifies how the border condition should be treated. The options are {"pad_outside", "pad_inside", "roll"}. "pad_outside" will offset the lattice in every direction by one step, and fill the new cells with the given `padding_value` and procedes to performing the computation; the resultant lattice in this case has the same shape as the initial lattice. "pad_inside" will perform the computation on the lattice, offsets inside by one cell from each side and returns the remainder cells; the resultant lattice is 2 cell smaller in each dimension than the original lattice. "roll" will assume that the end of each dimension is connected to the beginning of it and interprets the connectivity of the lattice with a rolling approach; the resultant lattice has the same shape is the original lattice. defaults to "pad_outside"
            padding_value (int, optional): 
                value used for padding in case the `border_condition` is set to "pad_outside"

        Raises:
            ValueError: 
                Main lattice and argument lattice shape should match
            NotImplementedError: 
                "pad_inside" is not implemented yet

        Returns:
            topogenesis.Lattice: 
                a new lattice containing the result of the application of the stencil
        """
        if self.shape != arg_lattice.shape:
            raise ValueError(
                "Main lattice and argument lattice shape does not match")

        if border_condition == "pad_outside":
            # pad the volume with padding value in every direction
            padded_arr = np.pad(self, (1, 1),
                                mode='constant',
                                constant_values=(padding_value, padding_value))
            # convert to lattice
            self_padded = to_lattice(padded_arr,
                                     self.minbound - self.unit,
                                     unit=self.unit)
            # pad the argument lattice with padding value in every direction
            padded_arg_arr = np.pad(arg_lattice, (1, 1),
                                    mode='constant',
                                    constant_values=(padding_value, padding_value))
            # convert to lattice
            arg_lattice_padded = to_lattice(padded_arg_arr,
                                            self.minbound - self.unit,
                                            unit=self.unit)

        elif border_condition == "pad_inside":
            raise NotImplementedError

        elif border_condition == "roll":
            self_padded = to_lattice(np.copy(self), self)
            arg_lattice_padded = to_lattice(np.copy(arg_lattice), arg_lattice)

        # find the neighbours based on the stencil
        cell_neighbors = self_padded.find_neighbours(stencil)

        # replace neighbours by their value in the main lattice
        self_padded_flat = self_padded.ravel()
        neighbor_values = self_padded_flat[cell_neighbors]

        # apply the function to the neighbour values
        applied = stencil.function(neighbor_values, axis=1)
        row_ind = np.arange(applied.size)

        # replace neighbours by their value in the argument latice
        arg_lattice_padded_flat = arg_lattice_padded.ravel()
        arg_neighbor_values = arg_lattice_padded_flat[cell_neighbors]

        # retrieve the values from the argument lattice
        arg_applied = arg_neighbor_values[row_ind, applied]

        # reshape the neighbour applied into the original lattice shape
        arg_applied_3d = arg_applied.reshape(self_padded.shape)

        # reverse the padding procedure
        if border_condition == "pad_outside":
            # trim the padded dimensions
            arg_applied_3d_trimmed = arg_applied_3d[1:-1, 1:-1, 1:-1]

        elif border_condition == "pad_inside":
            raise NotImplementedError

        elif border_condition == "roll":
            arg_applied_3d_trimmed = arg_applied_3d

        return to_lattice(arg_applied_3d_trimmed, self)

    def find_neighbours(self, stencil, order: str = "dist"):
        """Given an stencil, this method will return the neighbours of all cells with regard to the stencil specification in a numpy 2D array with each row corresponding to one cell in lattice.

        Args:
            stencil (topogenesis.Stencil): Stencil that describes the neighbourhood
            order (str, optional): the order of neighbours is one of {"dist", "C", "F"}. 'dist' sorts the neighbours based on the distance from origin cell, ‘C’ sorts in row-major (C-style) order, and ‘F’ sorts in column-major (Fortran- style) order. defaults to "dist"

        Returns:
            numpy.ndarray:  2D array describing the neighbours of each cell in the row
        """
        # TODO: kwarg for returning 1d or 3d indices
        # the id of voxels (0,1,2, ... n)
        self_ind = self.indices

        # calculating all the possible shifts to apply to the array
        shifts = stencil.expand(order)

        # gathering all the replacements in the columns
        replaced_columns = [
            np.roll(self_ind, shift, np.arange(3)).ravel() for shift in shifts]

        # stacking the columns
        cell_neighbors = np.stack(replaced_columns, axis=-1)

        return cell_neighbors

    def find_neighbours_masked(self, stencil, loc, order="dist", mask=None, border_condition="standard", id_type="1D"):
        """Given an stencil, this method will return the neighbours of one specific cell (specified with loc) with regard to the neighbourhood that stencil defines. The neighbours can be returned with 1D or 3D indices

        Args:
            stencil (topogenesis.Stencil): Stencil that describes the neighbourhood
            loc (numpy.ndarray): The location (3D index) of the cells that it's neighbours are desired.
            order (str, optional): the order of neighbours is one of {"dist", "C", "F"}. 'dist' sorts the neighbours based on the distance from origin cell, ‘C’ sorts in row-major (C-style) order, and ‘F’ sorts in column-major (Fortran- style) order. defaults to "dist"
            mask (numpy.ndarray, optional): Not implemented yet. Defaults to None.
            border_condition (str, optional): 
                specifies how the border condition should be treated. The options are {"standard", "roll"}. "standard" will assume that the cells at the border of bound of the lattice and they have less neighbours compared to the cells in the middle of the lattice. "roll" will assume that the end of each dimension is connected to the beginning of it and interprets the connectivity of the lattice with a rolling approach; the resultant lattice has the same shape is the original lattice. defaults to "standard"
            id_type (str, optional): specifies how the neighbours are specified and returened. The options are {"1D", "3D"}. "1D" will return a one-dimensional index of the cell, similar to the index that is assigned to each cell when lattice.indices is called. "3D" will return the three-dimensional index of the cell which is the integer coordinates of the cell as well. Defaults to "1D".

        Raises:
            NotImplementedError: in case that "roll" option is chosen for "border_condition"
            NotImplementedError: in case that "mask" option is used

        Returns:
            numpy.ndarray: array of the indices of the neighbours.
        """
        # the id of voxels (0,1,2, ... n)
        self_ind = self.indices

        # find the bounds of window around the specified location
        win_min = loc + stencil.minbound
        win_max = loc + stencil.maxbound + 1
        if border_condition == "standard":
            # ensure win_min is not less than zero
            win_min = np.maximum(win_min, np.array([0, 0, 0]))
            # ensure win_max is not more than shape -1
            win_max = np.minimum(win_max, np.array(self_ind.shape))

        # TODO
        if border_condition == "roll":
            raise NotImplementedError

        # TODO
        if mask != None:
            raise NotImplementedError

        self_ind = self_ind[win_min[0]: win_max[0],
                            win_min[1]: win_max[1],
                            win_min[2]: win_max[2]]

        # TODO: Sometimes, self_ind turns to be zero. need to be debugged
        # find the new 1D-index of the location
        new_loc_ind = np.ravel_multi_index(
            tuple(stencil.origin), self_ind.shape)

        # calculating all the possible shifts to apply to the array
        shifts = stencil.expand(order)

        # gathering all the replacements in the columns
        replaced_columns = [
            np.roll(self_ind, shift, np.arange(3)).ravel() for shift in shifts]

        # stacking the columns
        cell_neighbors = np.stack(replaced_columns, axis=-1)

        # extract 1D ids
        neighs_1d_id = cell_neighbors[new_loc_ind]

        if id_type == "1D":
            return neighs_1d_id

        elif id_type == "3D":
            # convert 1D index to 3D index
            neigh_3d_id = np.array(
                np.unravel_index(neighs_1d_id, self.shape)).T
            return neigh_3d_id


class cloud(np.ndarray):
    """ Cloud is a subclass of NumPy ndarrays to represent pointclouds in a continuous 
    3dimensional space. Clouds add spatial properties and functionalities such
    as bounds and voxelate]
    """

    def __new__(subtype, point_array, dtype=float, buffer=None, offset=0,
                strides=None, order=None):

        # extracting the shape from point_array
        shape = point_array.shape
        # using the point_array as the buffer
        buffer = point_array.flatten(order="C")

        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to cloud.__array_finalize__
        obj = super(cloud, subtype).__new__(subtype, shape, dtype,
                                            buffer, offset, strides,
                                            order)

        # set the  'bounds' attribute
        obj.bounds = np.array([obj.min(axis=0), obj.max(axis=0)])

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(cloud, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. cloud():
        #    obj is None
        #    (we're in the middle of the cloud.__new__
        #    constructor, and self.bounds will be set when we return to
        #    cloud.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(cloud):
        #    obj is arr
        #    (type(obj) can be cloud)
        # From new-from-template - e.g cloud[:3]
        #    type(obj) is cloud
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'bounds', because this
        # method sees all creation of default objects - with the
        # cloud.__new__ constructor, but also with
        # arr.view(cloud).
        self.bounds = getattr(obj, 'bounds', None)
        # We do not need to return anything

    @property
    def minbound(self):
        """Real minimum bound of the point cloud

        Returns:
            numpy.ndarray: real minimum bound
        """
        return self.bounds[0]

    @property
    def maxbound(self):
        """Real maximum bound of the point cloud

        Returns:
            numpy.ndarray: real maximum bound
        """
        return self.bounds[1]

    @classmethod
    def from_mesh_vertices(cls, mesh_path: str):
        """ Extracts the vertices of a mesh as the point cloud

        Args:
            mesh_path (str): path to the mesh file

        Returns:
            topogenesis.Cloud: The point cloud including the vertices of the mesh
        """
        # load the mesh using pyvista
        pv_mesh = pv.read(mesh_path)

        # extract vertices
        vertices = np.array(pv_mesh.points).astype(np.float64)

        # return vertices as cloud
        return cls(vertices)

    def voxelate(self, unit, tol=1e-6, **kwargs):
        """will voxelate the pointcloud based on a given unit size and returns a boolean lattice that describes which cells of the discrete space contained at least one point 

        Args:
            unit (int, numpy.array): describes the cell size of the resultant discrete space

        Raises:
            ValueError: if the size of the unit array does not correspond to 
        the number of dimensions in point cloud

        Returns:
            topogenesis.Lattice: boolean lattice describing which cells has contained at least one point
        """

        unit = np.array(unit)
        if unit.size != 1 and unit.size != self.minbound.size:
            raise ValueError(
                'the length of unit array needs to be either 1 or equal to the dimension of point cloud')
        elif unit.size == 1:
            unit = np.tile(unit, self.minbound.shape)

        closed = kwargs.get('closed', False)

        if closed:
            # R3 to Z3 : finding the closest voxel to each point
            point_scaled = self / unit
            # shift the hit points in each 2-dimension (n in 1-axes)
            shifts = np.array(list(itertools.product([-1, 0, 1], repeat=3)))
            shifts_steps = np.sum(np.absolute(shifts), axis=1)
            chosen_shift_ind = np.argwhere(shifts_steps == 3).ravel()
            sel_shifts = shifts[chosen_shift_ind]

            vox_ind = [np.rint(point_scaled + s * tol)
                       for s in sel_shifts]
            vox_ind = np.vstack(vox_ind)

        else:
            vox_ind = np.rint(self / unit).astype(int)

        # removing repetitions
        unique_vox_ind = np.unique(vox_ind, axis=0).astype(int)

        # mapping the voxel indices to real space
        reg_pnt = unique_vox_ind * unit

        # initializing the volume
        l = lattice([self.minbound - unit, self.maxbound + unit], unit=unit,
                    dtype=bool, default_value=False)

        # map the indices to start from zero
        mapped_ind = unique_vox_ind - np.rint(l.minbound/l.unit).astype(int)

        # setting the occupied voxels to True
        l[mapped_ind[:, 0], mapped_ind[:, 1], mapped_ind[:, 2]] = True

        return l

    def fast_vis(self, plot, color="#66beed"):
        """Adds the pointcloud to a pyvista plotter and returns it. 
        It is mainly used to rapidly visualize the point cloud

        Args:
            plot (pyvista.Plotter): a pyvista plotter

        Returns:
            pyvista.Plotter: the same pyvista plotter containing points of the pointcloud

        **Usage Example:**
        ```python
        p = pyvista.Plotter()
        cloud.fast_vis(
        ```
        """

        # adding the original point cloud: blue
        plot.add_mesh(pv.PolyData(self),
                      color=color,
                      point_size=3,
                      render_points_as_spheres=True,
                      label="Point Cloud")

        return plot

    def fast_notebook_vis(self, plot, color="#66beed"):
        """Adds the pointcloud to a pyvista ITK plotter and returns it. ITK plotters are specifically used in notebooks to plot the geometry inside the notebook environment It is mainly used to rapidly visualize the content of the lattice for visual confirmation

        Args:
            plot (pyvista.PlotterITK): a pyvista ITK plotter

        Returns:
            pyvista.PlotterITK: pyvista ITK plotter containing lattice features such as cells, bounding box, cell centroids

        ```python
        p = pyvista.PlotterITK()
        cloud.fast_notebook_vis(p)
        ```
        """
        # adding the original point cloud: blue
        plot.add_points(pv.PolyData(self), color=color)

        return plot

    def to_csv(self, path, delimiter=","):
        np.savetxt(path, self, delimiter=delimiter)


class stencil(np.ndarray):
    """ Stencil is a subclass of NumPy ndarrays to represent neighbourhoods in a discrete 3dimensional space. Certain functions can be assigned to stencil to add functionalities similar to kernels to them. They also provide convenience functions for defining "Moore" or Von Neumann" neighbourhoods. Ufuncs can also be used to alter stencils (Addition, subtraction, etc)]
    """

    def __new__(subtype, point_array, ntype="Custom", origin=np.array([0, 0, 0]), function=None, dtype=int, buffer=None, offset=0,
                strides=None, order=None):

        # extracting the shape from point_array
        shape = point_array.shape
        # using the point_array as the buffer
        buffer = point_array.flatten(order="C")

        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to stencil.__array_finalize__
        obj = super(stencil, subtype).__new__(subtype, shape, dtype,
                                              buffer, offset, strides,
                                              order)

        # set the neighbourhood type
        obj.ntype = ntype
        # set the origin
        obj.origin = origin
        # set the  'bounds' attribute
        shape_arr = np.array(shape)
        obj.bounds = np.array([shape_arr * 0, shape_arr - 1]) - origin
        # set the function attribute
        obj.function = function
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(stencil, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. stencil():
        #    obj is None
        #    (we're in the middle of the stencil.__new__
        #    constructor, and self.bounds will be set when we return to
        #    stencil.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(stencil):
        #    obj is arr
        #    (type(obj) can be stencil)
        # From new-from-template - e.g stencil[:3]
        #    type(obj) is stencil
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'properties', because this
        # method sees all creation of default objects - with the
        # stencil.__new__ constructor, but also with
        # arr.view(stencil).
        self.bounds = getattr(obj, 'bounds', None)
        self.ntype = getattr(obj, 'ntype', None)
        self.origin = getattr(obj, 'origin', None)
        self.function = getattr(obj, 'function', None)

        # We do not need to return anything

    def __array_wrap__(self, array, context=None):

        temp = np.array(array)

        # checking if the array has any value other than 0, and 1
        np.place(temp, temp > 0.5, [1])
        np.place(temp, temp < 0.5, [0])

        return stencil(temp, ntype="custom", origin=self.origin)

    @property
    def minbound(self):
        return self.bounds[0]

    @property
    def maxbound(self):
        return self.bounds[1]

    def expand(self, order: str = "dist"):
        """will return the local address of each filled cell. This can be utilized to access the neighbours of a cell in lattice.

        Args:
            order (str, optional): [description]. Defaults to "dist".

        Returns:
            numpy.ndarray: array in shape of (n, 3) to describe the relative location of neighbours
        """
        # list the locations
        locations = self.origin - np.argwhere(self)

        # check the ordering method

        # 'dist' means to sort based on the distance from origin
        if order == "dist":
            # calculating the distance of each neighbour
            sums = np.abs(locations).sum(axis=1)
            # sorting to identify the main cell
            ordered = np.argsort(sums)

        # ‘F’ means to sort in column-major (Fortran- style) order
        elif order == "F":
            ordered = np.arange(self.size).reshape(self.shape).flatten('F')

        # ‘C’ means to sort in row-major (C-style) order
        elif order == "C":
            ordered = np.arange(self.size).reshape(self.shape).flatten('C')

        # sort and return
        return locations[ordered].astype(int)

    def set_index(self, index, value: int):
        """Sets the value of a cell in stencil via local indexing (based on the origin cell)

        Args:
            index (list, numpy.array): local address of the desired cell
            value (int): the desired value to be set one of {0, 1}

        Raises:
            ValueError: if the local address is non-existence or incompatible with the shape of stencil
        """

        ind = np.array(index) + self.origin
        if ind.size != 3:
            raise ValueError(" the index needs to have three components")
        self[ind[0], ind[1], ind[2]] = value


def create_stencil(type_str: str, steps: int, clip: int = None):
    """Creates a stencil based on predefined neighbourhoods such as "von_neumann" or "moore".

    Args:
        type_str (str): 
            one of {"von_neumann", "moore", "boolean_marching_cube"}
        steps (int): 
            {"von_neumann", "moore"} neighbourhoods are defined based on how many steps far from the origin cell should be included.
        clip (int, optional): 
            will clip the defined neighbourhood, for example ("von_neumann", step=1) describes the 6-neighbourhood of a cell in 3dimensional lattice, ("von_neumann", step=2, clip=1) describes the 18-neighbourhood, and ("moore", step=1) describes 26-neighbourhood, defaults to None

    Raises:
        ValueError: if the neighbourhood type is unknown

    Returns:
        topogenesis.Stencil: the stencil of the perscribed neighbourhood
    """
    # check if clip is specified. if it is not, set it to the steps
    if clip == None:
        clip = steps
    # von neumann neighborhood
    if type_str == "von_neumann":
        # https://en.wikipedia.org/wiki/Von_Neumann_neighborhood

        # computing all the possible shifts to apply to the array
        shifts = np.array(list(itertools.product(
            list(range(-clip, clip+1)), repeat=3)))

        # the number of steps that the neighbour is appart from
        # the cell (step=1 : 6 neighbour, step=2 : 18 neighbours,
        # step=3 : 26 neighbours)
        shift_steps = np.sum(np.absolute(shifts), axis=1)

        # check the number of steps
        chosen_shift_ind = np.argwhere(shift_steps <= steps).ravel()

        # select the valid indices from shifts variable,
        # transpose them to get
        # separate indices in rows, add the number of
        # steps to make this an index
        locs = np.transpose(shifts[chosen_shift_ind]) + clip

        # initialize the stencil
        s = np.zeros((clip*2+1, clip*2+1, clip*2+1)).astype(int)

        # fill in the stencil
        s[locs[0], locs[1], locs[2]] = 1

        return stencil(s,
                       ntype=type_str,
                       origin=np.array([clip, clip, clip]))

    elif type_str == "moore":
        # https://en.wikipedia.org/wiki/Moore_neighborhood

        # computing all the possible shifts to apply to the array
        shifts = np.array(list(itertools.product(
            list(range(-clip, clip+1)), repeat=3)))

        # the number of steps that the neighbour is appart from the origin cell
        shift_steps = np.max(np.absolute(shifts), axis=1)

        # check the number of steps
        chosen_shift_ind = np.argwhere(shift_steps <= steps).ravel()

        # select the valid indices from shifts variable,
        # transpose them to get separate indices in rows,
        # add the number of steps to make this an index
        locs = np.transpose(shifts[chosen_shift_ind]) + clip

        # initialize the stencil
        s = np.zeros((clip*2+1, clip*2+1, clip*2+1)).astype(int)

        # fill in the stencil
        s[locs[0], locs[1], locs[2]] = 1

        return stencil(s, ntype=type_str, origin=np.array([clip, clip, clip]))

    elif type_str == "boolean_marching_cube":

        # initialize the stencil
        s = np.ones((2, 2, 2)).astype(int)

        return stencil(s, ntype=type_str, origin=np.array([0, 0, 0]))

    else:
        raise ValueError(
            'non-valid neighbourhood type for stencil creation')


def to_lattice(a, minbound: np.ndarray, unit=1, orient=np.array([0., 0., 0., 1.])) -> lattice:
    """Converts a numpy array into a lattice

    Args:
        a (numpy.ndarray): 
            array
        minbound (numpy.ndarray): 
            describing the minimum bound of the lattice in continuous space
        unit (int, optional): 
            the unit size of the lattice

    Returns:
        topogenesis.Lattice: the lattice representation of the array
    """
    # check if the minbound is a lattice
    if type(minbound) is lattice:
        l = minbound
        unit = l.unit
        minbound = l.minbound
        orient = l.orient

    # check if minbound is an np array
    elif type(minbound) is not np.ndarray:
        minbound = np.array(minbound)

    # compute the bounds based on the minbound
    bounds = np.array([minbound, minbound + unit * (np.array(a.shape) - 1)])

    # construct a lattice
    array_lattice = lattice(bounds, unit=unit, dtype=a.dtype, orient=orient)
    array_lattice[:, :, :] = a[:, :, :]
    return array_lattice





"""
Geometric Intersection Algorithms 
"""

import numpy as np
import pyvista as pv
import itertools
import concurrent.futures
import warnings
import os

from ..datastructures import cloud

__author__ = "Shervin Azadi, and Pirouz Nourian"
__copyright__ = "???"
__credits__ = ["Shervin Azadi", "Pirouz Nourian"]
__license__ = "???"
__version__ = "0.0.2"
__maintainer__ = "Shervin Azadi"
__email__ = "shervinazadi93@gmail.com"
__status__ = "Dev"

file_directory = os.path.dirname(os.path.abspath(__file__))


def mesh_sampling(mesh, unit, tol=1e-06, **kwargs):
    """This algorithm samples a mesh based on unit size

    Args:
        geo_mesh ([COMPAS Mesh]): [description]
        unit ([numpy array]): [Unit represents the unit size in the 
        sampling grid. It needs to be one float value or an array-like
        with three elements. In case that a scalar is given it will 
        used for all three dimensions]
        tol ([type], optional): [description]. Defaults to 1e-06.

    Returns:
        [type]: [description]
    """
    ####################################################
    # INPUTS
    ####################################################

    unit = np.array(unit)
    if unit.size == 1:
        unit = np.array([unit, unit, unit])
    elif unit.size != 3:
        raise ValueError(
            """unit needs to have three elements representing
            the unit size for mesh sampling in three dimensions""")
    dim_num = unit.size
    multi_core_process = kwargs.get('multi_core_process', False)
    return_ray_origin = kwargs.get('return_ray_origin', False)
    return_ray_dir = kwargs.get('return_ray_dir', False)

    # compare voxel size and tolerance and warn if it is not enough
    if min(unit) * 1e-06 < tol:
        warnings.warn(
            """Warning! The tolerance for rasterization is not small 
            enough, it may result in faulty results or failure of
            rasterization.Trydecreasingthetolerance or scaling
            the geometry.""")

    ####################################################
    # Initialize the volumetric array
    ####################################################
    mesh_vertices, mesh_faces = mesh

    # retrieve the bounding box information
    mesh_bb_min = np.amin(mesh_vertices, axis=0)
    mesh_bb_max = np.amax(mesh_vertices, axis=0)
    mesh_bb_size = mesh_bb_max - mesh_bb_min

    # find the minimum index in discrete space
    mesh_bb_min_z3 = np.rint(mesh_bb_min / unit).astype(int)
    # calculate the size of voxelated volume
    vol_size = np.ceil((mesh_bb_size / unit)+1).astype(int)
    # initiate the 3d array of voxel space called volume
    vol = np.zeros(vol_size)

    ####################################################
    # compute the origin and direction of rays
    ####################################################

    # increasing the vol_size by one to accommodate for
    # shooting from corners
    vol_size_off = vol_size + 1

    # retrieve the voxel index for ray origins
    hit_vol_ind = np.indices(vol_size_off)
    vol_ind_trans = np.transpose(hit_vol_ind) + mesh_bb_min_z3
    hit_vol_ind = np.transpose(vol_ind_trans)

    # retrieve the ray origin indices
    ray_orig_ind = [np.take(hit_vol_ind, 0, axis=d + 1)
                    .transpose((1, 2, 0))
                    .reshape(-1, 3)
                    for d in range(dim_num)]
    ray_orig_ind = np.vstack(ray_orig_ind)

    # retrieve the direction of ray shooting for each origin point
    normals = np.identity(dim_num).astype(int)

    # tile(stamp) the X-ray direction with the (Y-direction
    # * Z-direction) . Then repeat this for all dimensions
    # TODO: this line has a problem given the negative indices
    # are included now
    ray_dir = [np.tile(normals[d],
                       (vol_size_off[(d + 1) % dim_num]
                        * vol_size_off[(d + 2) % dim_num], 1))
               for d in range(dim_num)]
    ray_dir = np.vstack(ray_dir)

    # project the ray origin + shift it with half of the voxel size
    # to move it to corners of the voxels
    ray_orig = ray_orig_ind * unit + unit * -0.5  # * (1 - ray_dir)

    # project the ray origin
    proj_ray_orig = ray_orig * (1 - ray_dir)

    ####################################################
    # intersection
    ####################################################

    samples = []

    # check if multiprocessing is allowed
    if multi_core_process:
        # open the context manager
        with concurrent.futures.ProcessPoolExecutor() as executor:
            # submit the processes
            # results = [executor.submit(intersect, geo_mesh, face,
            # unit, mesh_bb_size, ray_orig, proj_ray_orig, ray_dir,
            # tol) for face in geo_mesh.faces()]
            results = [executor.submit(intersect, mesh_vertices[face], unit,
                                       mesh_bb_size, ray_orig, proj_ray_orig,
                                       ray_dir, tol) for face in mesh_faces]
            # fetch the results
            for f in concurrent.futures.as_completed(results):
                samples.extend(f.result())
    else:
        # iterate over the faces
        # for face in geo_mesh.faces():
        for face in mesh_faces:
            # print(face)
            # print(type(face))
            # id = np.array(face)
            # print(mesh_vertices[id])
            # print(mesh_vertices[np.array(face).astype(int)])
            # print(mesh_vertices[np.array(face)])

            face_hit_pos = intersect(mesh_vertices[face], unit, mesh_bb_size,
                                     ray_orig, proj_ray_orig, ray_dir, tol)
            samples.extend(face_hit_pos)

    ####################################################
    # OUTPUTS
    ####################################################

    # set the output list
    if len(samples) != 0:
        out = [cloud(np.array(samples))]
    else:
        out = [None]
    # check if the ray origins are requested
    if return_ray_origin:
        out.append(cloud(np.array(ray_orig)))
    # check if the ray direction are requested
    if return_ray_dir:
        out.append(ray_dir)

    # if the list has more than one item, return it as a tuple, if it
    # has only one item, return the item itself
    return tuple(out) if len(out) > 1 else out[0]


def intersect(face_vertices_xyz, unit, mesh_bb_size, ray_orig,
              proj_ray_orig, ray_dir, tol):
    face_hit_pos = []

    # check if the face is a triangle
    if face_vertices_xyz.shape[0] != 3:
        return([])

    # check if any coordinate of the projected ray origin is in betwen
    # the max and min of the coordinates of the face
    min_con = proj_ray_orig >= np.amin(
        face_vertices_xyz, axis=0)*(1 - ray_dir)
    max_con = proj_ray_orig <= np.amax(
        face_vertices_xyz, axis=0)*(1 - ray_dir)
    in_range_rays = np.all(min_con * max_con, axis=1)

    # retrieve the ray indices that are in range
    in_rang_ind = np.argwhere(in_range_rays).flatten()

    # iterate over the rays
    for ray in in_rang_ind:
        # retrieve ray direction
        direction = ray_dir[ray]
        # retrieve ray origin
        orig_pos = ray_orig[ray]
        # calc the destination of ray (max distance that it needs to
        # travel) this line has a problem given the negative indices
        # are included now
        # plus unit added in case of flat meshes
        dest_pos = orig_pos + direction * (mesh_bb_size + unit)

        # intersection
        # Translated from Pirouz C#
        hit_pt = triangle_line_intersect(
            (orig_pos, dest_pos), face_vertices_xyz, tol=tol)
        if hit_pt is not None:
            face_hit_pos.append(hit_pt)

    return(face_hit_pos)


def triangle_line_intersect(L, Vx, tol=1e-06):
    """
    Computing the intersection of a line with a triangle
    Algorithm from http://geomalgorithms.com/a06-_intersect-2.html
    C# implementation from https://github.com/Pirouz-Nourian/Topological_Voxelizer_CSharp/blob/master/Voxelizer_Functions.cs

    Args:
        L ([2d np array]): List of two points specified by their coordinates
        Vx ([2d np array]): List of three points specified by their coordinates
        tol ([float], optional): tolerance. Defaults to 1e-06.

    Raises:
        ValueError: If the triangle contains more than three vertices

    Returns:
        [np array]: [description]
    """

    ####################################################
    # INPUTS
    ####################################################

    if len(Vx) != 3:
        raise ValueError('A triangle needs to have three vertexes')

    ####################################################
    # PROCEDURE
    ####################################################

    # finding U & V vectors
    O = Vx[0]
    U = Vx[1] - Vx[0]
    V = Vx[2] - Vx[0]
    # finding normal vector
    N = surface_normal_newell_vectorized(Vx)  # np.cross(U, V)

    Nomin = np.dot((O - L[0]), N)
    Denom = np.dot(N, (L[1] - L[0]))

    if Denom != 0:
        alpha = Nomin / Denom

        # L[0] + np.dot(alpha, (L[1] - L[0])): parameter along the
        # line where it intersects the plane in question, only if
        # not parallel to the plane
        W = L[0] + np.dot(alpha, (L[1] - L[0])) - O

        UU = np.dot(U, U)
        VV = np.dot(V, V)
        UV = np.dot(U, V)
        WU = np.dot(W, U)
        WV = np.dot(W, V)

        STDenom = UV**2 - UU * VV

        s = (UV * WV - VV * WU) / STDenom
        t = (UV * WU - UU * WV) / STDenom

    ####################################################
    # OUTPUTS
    ####################################################

        if s + tol >= 0 and t + tol >= 0 and s + t <= 1 + 2*tol:
            Point = O + s * U + t * V
            return Point
        else:
            return None
    else:
        return None


def surface_normal_newell_vectorized(poly):
    """    
    https://stackoverflow.com/questions/39001642/calculating-surface-normal-in-python-using-newells-method
    Newell Method explained here: https://www.researchgate.net/publication/324921216_Topology_On_Topology_and_Topological_Data_Models_in_Geometric_Modeling_of_Space

    Args:
        poly ([2d np array]): List of vertices specified by their coordinates 

    Raises:
        ValueError: [description]

    Returns:
        [type]: [description]
    """

    # This section is the vectorized equivalent of this code

    """
    n = np.array([0.0, 0.0, 0.0])

    for i in range(3):
        j = (i+1) % 3
        n[0] += (poly[i][1] - poly[j][1]) * (poly[i][2] + poly[j][2])
        n[1] += (poly[i][2] - poly[j][2]) * (poly[i][0] + poly[j][0])
        n[2] += (poly[i][0] - poly[j][0]) * (poly[i][1] + poly[j][1])
    """
    poly_10 = np.roll(poly, [-1, 0], np.arange(2))
    poly_01 = np.roll(poly, [0, -1], np.arange(2))
    poly_11 = np.roll(poly, [-1, -1], np.arange(2))

    n = np.roll(np.sum((poly - poly_10) * (poly_01 + poly_11), axis=0), -1, 0)

    norm = np.linalg.norm(n)
    if norm == 0:
        raise ValueError('zero norm')
    else:
        normalized = n/norm

    return normalized




"""
Cloud DataStructure
"""

import numpy as np
import pandas as pd
import pyvista as pv
import itertools
import concurrent.futures
import warnings
import os

from .datastructures import stencil, cloud, lattice, to_lattice

__author__ = "Shervin Azadi, and Pirouz Nourian"
__copyright__ = "???"
__credits__ = ["Shervin Azadi", "Pirouz Nourian"]
__license__ = "???"
__version__ = "0.0.2"
__maintainer__ = "Shervin Azadi"
__email__ = "shervinazadi93@gmail.com"
__status__ = "Dev"

file_directory = os.path.dirname(os.path.abspath(__file__))


def scatter(bounds, count):

    point_array = np.random.uniform(bounds[0],
                                    bounds[1],
                                    (count, bounds.shape[1]))
    return cloud(point_array)


def cloud_from_csv(file_path, delimiter=','):

    point_array = np.genfromtxt(file_path, delimiter=delimiter)
    return cloud(point_array)


def cloud_from_mesh_vertices(mesh_path):

    # load the mesh using pyvista
    pv_mesh = pv.read(mesh_path)

    # extract vertices
    v = np.array(pv_mesh.points)

    # return vertices as cloud
    return cloud(v)


# extra import function
# TODO: this function is updated make sure that it is functioning properly
def lattice_from_csv(file_path):
    # read metadata
    meta_df = pd.read_csv(file_path, nrows=3)

    shape = np.array(meta_df['shape'])
    unit = np.array(meta_df['unit'])
    minbound = np.array(meta_df['minbound'])

    # read lattice
    lattice_df = pd.read_csv(file_path, skiprows=5)

    # create the buffer
    buffer = np.array(lattice_df['value']).reshape(shape)

    # create the lattice
    l = to_lattice(buffer, minbound=minbound, unit=unit)

    return l


def marching_cube_vis(p, cube_lattice, style_str):

    # extract cube indices
    cube_ind = np.transpose(np.indices(cube_lattice.shape),
                            (1, 2, 3, 0)).reshape(-1, 3)
    # extract cube positions
    cube_pos = (cube_ind + 0.5) * cube_lattice.unit + cube_lattice.minbound

    # extract cube tid
    cube_tid = cube_lattice.ravel()

    # remove the cube position and tid where tid is 0
    filled_cube_pos = cube_pos[cube_tid > 0]
    filled_cube_tid = cube_tid[cube_tid > 0]

    if style_str != "chamfer":
        raise ValueError(
            "Meshing style is not valid. Valid styles are: ['chamfer']")

    # load tiles
    tiles = [0]
    for i in range(1, 256):
        tile_path = os.path.join(os.path.dirname(
            file_directory), "resources/mc_tiles", style_str, "Tile_{0:03d}.obj".format(i))
        tile = pv.read(tile_path)
        tile.points *= cube_lattice.unit
        tiles.append(tile)

    new_points = tiles[filled_cube_tid[0]].points + filled_cube_pos[0]
    new_faces = tiles[filled_cube_tid[0]].faces.reshape(-1, 4)

    # merge tiles
    for i in range(1, filled_cube_tid.size):
        tile = tiles[filled_cube_tid[i]]
        # add the faces list, changing the point numbers
        new_faces = np.concatenate(
            (new_faces, tile.faces.reshape(-1, 4) + np.array([0, 1, 1, 1])*new_points.shape[0]), axis=0)
        # add the new points, change the position based on the location
        new_points = np.concatenate(
            (new_points, tile.points + filled_cube_pos[i]), axis=0)

    # construct the new mesh and add it to plot
    new_tile = pv.PolyData(new_points, new_faces)
    p.add_mesh(new_tile, color='#abd8ff')

    return p




"""
Stencil Functions
"""

import numpy as np
import warnings
import os

file_directory = os.path.dirname(os.path.abspath(__file__))

# directly using numpy functions
sum = np.sum
min = np.amin
argmin = np.argmin

# wrapping numpy functions to enable mapping


def random_choice(array, axis=1):
    return np.apply_along_axis(np.random.choice, axis, array)



from setuptools import setup, find_packages

setup(name='topogenesis',
      version='0.0.5',
      description='Topological Structures and Methods for Generative Systems and Sciences',
      url='https://github.com/shervinazadi/topoGenesis',
      author='Shervin Azadi, and Pirouz Nourian',
      author_email='shervinazadi93@gmail.com',
      
      license='MIT',
      packages=find_packages(),
      zip_safe=False,
      python_requires='>=3.7',
      include_package_data=True,
      install_requires=[
    'numpy',
    'pandas',
    # any other dependencies
],
      )





"""
Geometry loading modules of topoGenesis
"""

import numpy as np
import pyvista as pv
import warnings
import os

__author__ = "Shervin Azadi, and Pirouz Nourian"
__copyright__ = "???"
__credits__ = ["Shervin Azadi", "Pirouz Nourian"]
__license__ = "???"
__version__ = "0.0.2"
__maintainer__ = "Shervin Azadi"
__email__ = "shervinazadi93@gmail.com"
__status__ = "Dev"

file_directory = os.path.dirname(os.path.abspath(__file__))


def load_mesh(mesh_path, check_is_triangle=True):
    # load the mesh using pyvista
    pv_mesh = pv.read(mesh_path)

    # check if all of the faces are triangles
    if check_is_triangle and not pv_mesh.is_all_triangles():
        raise ValueError('All faces need to be triangles!')

    # extract faces and vertices
    v = np.array(pv_mesh.points)
    f = pv_mesh.faces.reshape(-1, 4)[:, 1:]

    # return them as 3d numpy arrays
    return np.array(v).astype(np.float64), np.array(f).astype(np.int64)
